unit DD92u1;

(**)

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls;

type
  TDD92f1 = class(TForm)
    buDoIt: TButton;
    buQuit: TButton;
    laQuote: TLabel;
    laSourceOfQuote: TLabel;
    Memo1: TMemo;
    procedure buQuitClick(Sender: TObject);
    procedure buDoItClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
  private
    { Private declarations }
    sTmp,sTmp2,sQuote,sSourceOfQuote,sQuotesLibFilename:string;
    boTmp:boolean;
    bTmp:byte;
    procedure GetQuoteAndSource;
    procedure FillMemo;
    procedure MassageAndVett;

  public
    { Public declarations }
  end;

var
  DD92f1: TDD92f1;

implementation

{$R *.DFM}

procedure TDD92f1.buQuitClick(Sender: TObject);
begin
application.terminate;
end;

procedure TDD92f1.buDoItClick(Sender: TObject);
begin
GetQuoteAndSource;
laQuote.caption:=sQuote;
laSourceOfQuote.caption:=sSourceOfQuote;
end;

procedure TDD92f1.FormCreate(Sender: TObject);
begin
laQuote.caption:='Welcome to the Quote Machine, DD92';
laSourceOfQuote.caption:='Delphi Tutorial from Sheepdog Software';
FillMemo;
MassageAndVett;
end;

procedure TDD92f1.GetQuoteAndSource;
var c1:integer;
begin
//Pick a line as a possible quote... and
//  go to the next line if it isn't one.
c1:=random(memo1.lines.count);
repeat
  sQuote:=memo1.lines[c1];
  inc(c1);
until copy(sQuote,1,2)='qq';

//Fill sSourceOfQuote... with 'qs' if there
//  is no source given in the datafile.
sSourceOfQuote:='qs';
sTmp:=memo1.lines[c1];
if sTmp[2]='s' then sSourceOfQuote:=sTmp;

//Strip off the 'qq' and 'qs' at the starts
//  of the quote and source....
sQuote:=copy(sQuote,3,length(sQuote)-2);
sSourceOfQuote:=copy(sSourceOfQuote,3,length(sSourceOfQuote)-2);

end;//GetQuoteAndSource

{procedure TDD92f1.FillMemo;
begin
Memo1.lines.LoadFromFile('DD92quotes.txt');
end;}

procedure TDD92f1.FillMemo;
begin
sQuotesLibFilename:='DD92quotes.txt';
if fileexists(sQuotesLibFilename) then
Memo1.lines.LoadFromFile(sQuotesLibFilename)//no ; here
else begin
   showmessage('A text file named '+sQuotesLibFilename+' must '+
   'exist in the same folder as the application''s .exe file. '+
   'It must consist of at least two lines conforming to '+
   'the following pattern (quote and source).....'+chr($d)+chr($d)+
   '   <!--q-->E tu Brute?'+chr($d)+
   '   <!--s-->Caesar, J'+chr($d)+chr($d)+
   'Provide such a file.'+chr($d)+
   'Program will shut down after you press OK.');
   application.terminate;
   end;//else
end;


procedure TDD92f1.MassageAndVett;
var c1,c2:integer;//Must be able to hold -1
begin
(*This is a long procedure, but basically quite
  pedestrian. It just goes through the data in the
  memo, appliying certain massages, and making
  certain checks that the data is suitably arranged
  for the needs of the Quote Machine.

  Beware "fencepost" problems. First line is line [0], and the
  memo1.lines.count will equal FIVE when there are FIVE lines,
  but the last line will be line[4]!

  Beware also: If, with a c1 that means you are working
  with the then-last line of the memo, you use...

  memo1.lines[c1]:=sTmp;
  or
  memo1.lines.add(sTmp);

  ... then it may SEEM that an blank EXTRA line is added to the memo,
  because you can now use the down arrow to move onto the next
  line, whereas previously you could not... but the count of
  lines in the memo has not gone up. A puzzle, and one that can
  drive you crazy when debugging "last line" questions. Sigh.

  (These notes made while working with Delphi 4.)
*)

//Remove leading and trailing spaces and control characters...
c1:=0;
c2:=memo1.lines.count;
repeat
  sTmp:=memo1.lines[c1];
  if sTmp<>'' then begin
       if (sTmp[1]=' ') then begin
          sTmp:=trim(sTmp);//strips off leading and trailing spaces and control characters.
          //N.B.: We are only looking to strip LEADING spaces... lines with just TRAILING
          //   spaces... or starting with a control character... will not be modified.
          memo1.lines[c1]:=sTmp;
       end;//if sTmp[1]=' '...
     end;//if sTmp[1]=<>''...
  inc(c1);
until c1=c2;
(*Yes... the above WILL remove leading spaces from last line,
and the "until" is satisfied when c1 is "too big". Inc(c1)
happens AFTER attempt to work with current line[c1].
If there are 3 lines in the file, count will be 3, but the
last line will be line[2]*)

//Now....
//Remove blank lines, and lines starting with semicolons (;)...
//A line starting with a space and THEN a semicolon will also
//be removed, because spaces were removed before we looked
//for blank or semicolon lines.

c1:=0;
c2:=memo1.lines.count;
repeat
  sTmp:=memo1.lines[c1];
  if sTmp<>'' then if sTmp[1]=';' then sTmp:='';
  if sTmp='' then begin
     memo1.lines.delete(c1);
     dec(c1);
     dec(c2);
     end;
  inc(c1);
until c1>=c2;// '>' arises when you delete the last line.

//Now remove header lines...
c1:=0;
c2:=memo1.lines.count;
boTmp:=false;
repeat
  sTmp:=memo1.lines[c1];
  if copy(sTmp,1,8)='<!--q-->' then boTmp:=true;
  if boTmp=false then begin
     memo1.lines.delete(c1);
     dec(c1);
     dec(c2);
     end;
  inc(c1);
until (boTmp=true) or (c1=c2);

if c1=c2 then begin
  showmessage('Your datafile does not appear to have even '+
    'one line with <!--q--> at the start of a line. '+
    'The application will shut down. Revise your '+
    'datafile and try again.');
  application.terminate;
  //N.B. "application.terminate" used thus does not
  //always succeed, but it mostly will, and, I hope,
  //you won't try to run the app with an invalid datafile
  //anyway. But you should always TRY, as best you can,
  //to provide for the many, many possible user errors.
  //Thinking of ALL the mistakes users may make is the
  //hard part. Idiot-proofing things would be so much
  //easier if idiots were not so clever.
  end;

//Make sure that no line starts with "qq" or "qs"
//   at this point. (We haven't yet converted the
//   human friendly tags to the shorter tags.
bTmp:=255;
c1:=0;
c2:=memo1.lines.count-1;
repeat
  sTmp:=copy(memo1.lines[c1],1,2);
  if sTmp='qq' then bTmp:=0;
  if sTmp='qs' then bTmp:=1;
  inc(c1);
until (c1=c2) or (bTmp<>255);

if bTmp=0 then begin
   showmessage('There is at least one line in the data '+
   'file which, before massaging, starts with "qq". '+
   'This is not allowed. Revise your datafile. The '+
   'application will shut down when you click "OK".');
   application.terminate;
   end;

if bTmp=1 then begin
   showmessage('There is at least one line in the data '+
   'file which, before massaging, starts with "qs". '+
   'This is not allowed. Revise your datafile. The '+
   'application will shut down when you click "OK".');
   application.terminate;
   end;

//Convert "human friendly" tags to shorter tags, and
//  process any "qq"s or "qs"s which are supposed to start
//  a quote or the description of a quote's source...
c1:=0;
c2:=memo1.lines.count;
repeat
  bTmp:=255;//Biggest number you can put in a Delphi byte- type variable
  sTmp:=memo1.lines[c1];
  if copy(sTmp,1,8)='<!--q-->' then bTmp:=0;
  if copy(sTmp,1,8)='<!--s-->' then bTmp:=1;
  if copy(sTmp,1,2)='qq' then bTmp:=2;
  if copy(sTmp,1,2)='qs' then bTmp:=3;
  //Note: AFTER this process, there will be many lines
  //starting qq or qs. But IN this process, we convert
  //lines which STARTED starting with qq or qs to lines
  //starting q-q and q-s, respectively.

  if bTmp=0 then begin
    sTmp:='qq'+copy(sTmp,9,length(sTmp)-8);
    memo1.lines[c1]:=sTmp;
    end;
  if bTmp=1 then begin
    sTmp:='qs'+copy(sTmp,9,length(sTmp)-8);
    memo1.lines[c1]:=sTmp;
    end;
  if bTmp=2 then begin
    sTmp:='q-q'+copy(sTmp,3,length(sTmp)-2);
    memo1.lines[c1]:=sTmp;
    end;
  if bTmp=3 then begin
    sTmp:='q-s'+copy(sTmp,3,length(sTmp)-2);
    memo1.lines[c1]:=sTmp;
    end;

  inc(c1);
until c1=c2;

//Put all tags ("qq"/""qs") at the start of the data they
//   mark, i.e. don't leave them on their own lines...
c1:=0;
c2:=memo1.lines.count;
repeat
  sTmp:=memo1.lines[c1];
  if sTmp='qq' then begin
     sTmp:='qq'+memo1.lines[c1+1];
     memo1.lines[c1]:=sTmp;
     // and now take the extra line out...
     memo1.lines.delete(c1+1);
     dec(c1);
     dec(c2);
     end;

  if sTmp='qs' then begin
     sTmp:='qs'+memo1.lines[c1+1];
     memo1.lines[c1]:=sTmp;
     // and now take the extra line out...
     memo1.lines.delete(c1+1);
     dec(c1);
     dec(c2);
     end;

  inc(c1);
until c1=c2-1;
//minus 1 in case file ends with "qq" or "qs".... which it shouldn't... but
//  it could. That particular error will be caught in a moment, when we check
//  that the last line of the file is a "qs+source description" line.
//HERE, we are only conflating qq and qs lines which are on their own,
//  conflating them with what comes on the line after them.

//=========
//Check each line except the last starts with a qq or
//   a qs. (We check the last line later, anyway, and there
//   is a quirk (bug) in the program still which means that
//   the last line at this point is sometimes still a blank
//   line, even though that "shouldn't" happen. Sigh.
bTmp:=255;
c1:=0;
c2:=memo1.lines.count-1;
repeat
  sTmp:=copy(memo1.lines[c1],1,2);
  if (sTmp<>'qs') and (stmp<>'qq') then bTmp:=0;
  inc(c1);
until (c1=c2) or (bTmp<>255);

if bTmp=0 then begin
   showmessage('There is at least one line in the data '+
   'file which, after massaging, doesn''t begin with '+
   'either "qq" or "qs". '+
   'Revise your datafile. The application '+
   'will shut down when you click "OK".');
   application.terminate;
   end;

//=========
//Right... NOW we can check that the file ends with a line starting qs, and
//  having a source- of- quote description after it. Remember: Not EVERY
//  quote needs to have a source description, but the last one in the
//  file DOES need one, to save us some hassle that I described elsewhere.
//  (That hassle associated with USING the file of quotes)

//BLUSH: I had to add a little kludge here. There SHOULDN'T be a blank
//  line at the end of the memo by now... but it turns out, somehow, that
//  there can be one. IF there is, at this point, it cannot... I hope...
//  be a problem. Other things we've done so far should still be "happy"...
//  So it is "okay" (not ideal, but "okay") to just take off what should
//  not be here... and THEN make the check we were going to make...

c1:=memo1.lines.count;
repeat
sTmp:=memo1.lines[c1];
if sTmp='' then begin
  memo1.lines.delete(c1);
  dec(c1);
  sTmp:=memo1.lines[c1];
  end;
until sTmp<>'';
//Could we have a case where we entered that loop when there
//  was NOTHING but blank lines in the memo? Perhaps. And if
//  that arose, the program would crash. I'm going to take
//  THAT chance. Got to live dangerously once in a while.

//The sort of highly specific tests in the next line are
//  not always possible... but that doesn't mean that they
//  are NEVER possible. But when you use them, always ask
//  yourself... "Can I rely on things being like this?"
//  "Am I covering all the possibilities?"
//At this point in this program I know a lot, and I
//  only want to check one "detail"... so this works...
//  ... if I haven't overlooked something!

if(length(sTmp)<3) or (copy(sTmp,1,2)<>'qs') then begin
   showmessage('The last line of the data file must be '+
   'describing the source of a quote. Not EVERY quote '+
   'needs its source described, but the last quote DOES '+
   'need this. Revise your datafile. The application '+
   'will shut down when you click "OK".');
   application.terminate;
   end;

//Next, check that memo now consists only of qq/qs lines, with
//  no cases of two qs lines in a row. We don't need to check
//  that the last line is a qs line, but we do need to check
//  that the penultimate line wasn't also a qs line, and the
//  easiest way to do that is to make a special case out of
//  it....

//During these tests, we will also check that no line is
//  JUST "qq" or "qs", which would also be wrong.

bTmp:=255;//For now, assume that it is true that there
   //are no problems in the structure of the memo.

//check penultimate line...
sTmp:=memo1.lines[memo1.lines.count-2];
if length(sTmp)<3 then bTmp:=0;
if copy(sTmp,1,2)<>'qq' then bTmp:=1;

//check last line...
sTmp:=memo1.lines[memo1.lines.count-1];
if length(sTmp)<3 then bTmp:=0;
if copy(sTmp,1,2)<>'qs' then bTmp:=3;

//Now check all lines are at least 3 characters, and that
//  only qq lines follow qs lines. (In the body of the memo
//  two qq lines in a row are ok, and we've already checked
//  that the last quote record has both a qq and a qs part.

//OOPs... late in the day, I realized that one(?) more test
//  is needed: 1st line must be a qq line. Happily, because
//  of the structure of the program, adding the extra test
//  was not very difficult.


c1:=0;
c2:=memo1.lines.count-2;
repeat
  sTmp:=memo1.lines[c1];
    if length(sTmp)<3 then bTmp:=0;
    if copy(sTmp,1,2)='qs' then
       if copy(memo1.lines[c1+1],1,2)='qs' then bTmp:=2;
  inc(c1);
until (c1=c2) or (bTmp<>255);

if bTmp=0 then begin
   showmessage('There is at least one line in the data '+
   'file which, after massaging, is too short. '+
   'Revise your datafile. The application '+
   'will shut down when you click "OK".');
   application.terminate;
   end;

if bTmp=1 then begin
   showmessage('The penultimate line in the data '+
   'file, after massaging, must be a <!--q--> line, and '+
   'it isn''t. Revise your datafile. The application '+
   'will shut down when you click "OK".');
   application.terminate;
   end;

if btmp=2 then begin
   showmessage('You cannot, as you have now, '+
   'have two  <!--s--> lines in a row. '+
   'Revise your datafile. The application '+
   'will shut down when you click "OK".');
   application.terminate;
   end;

if bTmp=3 then begin
   showmessage('The last line in the data '+
   'file, after massaging, must be a <!--s--> line, and '+
   'it isn''t. Revise your datafile. The application '+
   'will shut down when you click "OK".');
   application.terminate;
   end;

if bTmp=4 then begin
   showmessage('The first line in the data '+
   'file, after massaging, including the removal '+
   'of the header lines, must be a <!--q--> line, and '+
   'it isn''t. Revise your datafile. The application '+
   'will shut down when you click "OK".');
   application.terminate;
   end;
end;//MassageAndVett

end.
